# -*- coding: utf-8 -*-
#
#       pointer.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.scribble` -- Manage user drawings on the current slide
---------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import math

import gi
import cairo
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

from pympress import builder, extras, util


class Scribbler(builder.Builder):
    """ UI that allows to draw free-hand on top of the current slide.

    Args:
        config (:class:`~pympress.config.Config`): A config object containing preferences
        builder (:class:`~pympress.builder.Builder`): A builder from which to load widgets
        notes_mode (`bool`): The current notes mode, i.e. whether we display the notes on second slide
    """
    #: Whether we are displaying the interface to scribble on screen and the overlays containing said scribbles
    scribbling_mode = False
    #: `list` of scribbles to be drawn, as tuples of color :class:`~Gdk.RGBA`, width `int`, and a `list` of points.
    scribble_list = []
    #: `list` of undone scribbles to possibly redo
    scribble_redo_list = []
    #: Whether the current mouse movements are drawing strokes or should be ignored
    scribble_drawing = False
    #: :class:`~Gdk.RGBA` current color of the scribbling tool
    scribble_color = Gdk.RGBA()
    #: `int` current stroke width of the scribbling tool
    scribble_width = 1

    #: :class:`~Gtk.HBox` that replaces normal panes when scribbling is on, contains buttons and scribble drawing area.
    scribble_overlay = None
    #: :class:`~Gtk.DrawingArea` for the scribbles in the Presenter window. Actually redraws the slide.
    scribble_p_da = None
    #: :class:`~Gtk.EventBox` for the scribbling in the Content window, captures freehand drawing
    scribble_c_eb = None
    #: :class:`~Gtk.EventBox` for the scribbling in the Presenter window, captures freehand drawing
    scribble_p_eb = None
    #: :class:`~Gtk.AspectFrame` for the slide in the Presenter's highlight mode
    scribble_p_frame = None
    #: The :class:`~Gtk.DrawingArea` in the content window
    c_da = None

    #: The :class:`~Gtk.ColorButton` selecting the color of the pen
    scribble_color_selector = None
    #: The :class:`~Gtk.Scale` selecting the size of the pen
    scribble_width_selector = None
    #: The `list` containing the radio buttons :class:`~Gtk.ModelButton`
    scribble_preset_buttons = []

    #: The position of the mouse on the slide as `tuple` of `float`
    mouse_pos = None
    #: A :class:`~cairo.Surface` to hold drawn highlights
    scribble_cache = None
    #: The next scribble to render (i.e. that is not rendered in cache)
    next_render = 0

    #: :class:`~Gtk.Button` for removing the last drawn scribble
    scribble_undo = None
    #: :class:`~Gtk.Button` for drawing the last removed scribble
    scribble_redo = None
    #: :class:`~Gtk.Button` for removing all drawn scribbles
    scribble_clear = None

    #: A :class:`~Gtk.OffscreenWindow` where we render the scribbling interface when it's not shown
    scribble_off_render = None
    #: :class:`~Gtk.Box` in the Presenter window, where we insert scribbling.
    p_central = None

    #: :class:`~Gtk.Button` that is clicked to stop zooming, unsensitive when there is no zooming
    zoom_stop_button = None

    #: callback, to be connected to :func:`~pympress.surfacecache.SurfaceCache.resize_widget`
    resize_cache = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.on_draw`
    on_draw = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.track_motions`
    track_motions = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.track_clicks`
    track_clicks = lambda *args: None

    #: callback, to be connected to :func:`~pympress.ui.UI.load_layout`
    load_layout = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.redraw_current_slide`
    redraw_current_slide = lambda *args: None

    #: callback, to be connected to :func:`~pympress.extras.Zoom.get_slide_point`
    get_slide_point = lambda *args: None
    #: callback, to be connected to :func:`~pympress.extras.Zoom.start_zooming`
    start_zooming = lambda *args: None
    #: callback, to be connected to :func:`~pympress.extras.Zoom.stop_zooming`
    stop_zooming = lambda *args: None

    #: `int` that is the currently selected element
    active_preset = -1
    #: `int` to remember the previously selected element, before holding “eraser”
    previous_preset = -1
    #: `list` that contains the modifiers which, when held on scribble start, toggle the eraser
    toggle_erase_modifiers = []
    #: `list` that contains the non-modifier shortcuts which, when held on scribble start, toggle the eraser
    toggle_erase_shortcuts = []
    #: `str` or `None` that indicates whether a modifier + click or a held shortcut is toggling the eraser
    toggle_erase_source = None

    #: The :class:`~Gio.Action` that contains the currently selected pen
    pen_action = None

    #: `str` which is the mode for scribbling, one of 3 possile values:
    # global means one set of scribbles for the whole document
    # single-page means we manage a single page of scribbles, and clear everything on page change (historical behaviour)
    # per-page means we manage a set of scribbles per document page, and clear or restore them on page change
    # per-label means we manage a set of scribbles per document page, but defined by label and not page number
    highlight_mode = 'single-page'
    #: `bool` indicating whether we exit highlighting mode on page change
    page_change_exits = True

    #: `dict` of scribbles per page
    remembered_scribbles = {}
    #: `tuple` of (`int`, `str`) indicating the current page number and label
    current_page = (None, None)

    #: `str` indicating the current layout of the highlight toolbar
    tools_orientation = 'vertical'
    #: :class:`~Gtk.Box` containing the presets
    preset_toolbar = None
    #: :class:`~Gtk.Box` containing the scribble buttons
    scribble_toolbar = None
    #: :class:`~Gtk.Box` containing the scribble color and width selectors
    scribble_color_toolbox = None


    def __init__(self, config, builder, notes_mode):
        super(Scribbler, self).__init__()

        self.load_ui('highlight')
        builder.load_widgets(self)
        self.get_application().add_window(self.scribble_off_render)

        self.on_draw = builder.get_callback_handler('on_draw')
        self.track_motions = builder.get_callback_handler('track_motions')
        self.track_clicks = builder.get_callback_handler('track_clicks')
        self.load_layout = builder.get_callback_handler('load_layout')
        self.redraw_current_slide = builder.get_callback_handler('redraw_current_slide')
        self.resize_cache = builder.get_callback_handler('cache.resize_widget')
        self.get_slide_point = builder.get_callback_handler('zoom.get_slide_point')
        self.start_zooming = builder.get_callback_handler('zoom.start_zooming')
        self.stop_zooming = builder.get_callback_handler('zoom.stop_zooming')

        self.connect_signals(self)
        self.config = config

        # Prepare cairo surfaces for markers, with 3 different marker sizes, and for eraser
        ms = [1, 2, 3]
        icons = [cairo.ImageSurface.create_from_png(util.get_icon_path('marker_{}.png'.format(n))) for n in ms]
        masks = [cairo.ImageSurface.create_from_png(util.get_icon_path('marker_fill_{}.png'.format(n))) for n in ms]

        self.marker_surfaces = list(zip(icons, masks))
        self.eraser_surface = cairo.ImageSurface.create_from_png(str(util.get_icon_path('eraser.png')))

        # Load color and active pen preferences. Pen 0 is the eraser.
        self.color_width = [(Gdk.RGBA(0, 0, 0, 0), 150)] + list(zip(
            [self.parse_color(config.get('highlight', 'color_{}'.format(pen))) for pen in range(1, 10)],
            [config.getint('highlight', 'width_{}'.format(pen)) for pen in range(1, 10)],
        ))

        self.scribble_preset_buttons = [
            self.get_object('pen_preset_{}'.format(pen) if pen else 'eraser') for pen in range(10)
        ]

        self.tools_orientation = self.config.get('layout', 'highlight_tools')
        self.adjust_tools_orientation()

        active_pen = config.get('highlight', 'active_pen')
        self.page_change_exits = config.getboolean('highlight', 'page_change_exits')
        self.setup_actions({
            'highlight':         dict(activate=self.switch_scribbling, state=False),
            'highlight-use-pen': dict(activate=self.load_preset, state=active_pen, parameter_type=str, enabled=False),
            'highlight-clear':   dict(activate=self.clear_scribble),
            'highlight-redo':    dict(activate=self.redo_scribble),
            'highlight-undo':    dict(activate=self.pop_scribble),
            'highlight-mode':    dict(activate=self.set_mode, state=self.highlight_mode, parameter_type=str),
            'highlight-page-exit': dict(activate=self.page_change_action, state=self.page_change_exits),
            'highlight-tools-orientation': dict(activate=self.set_tools_orientation, state=self.tools_orientation,
                                                parameter_type=str),
        })

        hold_erase = [Gtk.accelerator_parse(keys) for keys in config.shortcuts.get('highlight-hold-to-erase', [])]
        self.toggle_erase_modifiers = [mod for keycode, mod in hold_erase if not keycode]
        self.toggle_erase_shortcuts = [(keycode, mod) for keycode, mod in hold_erase if keycode]

        self.pen_action = self.get_application().lookup_action('highlight-use-pen')
        self.load_preset(self.pen_action, int(active_pen) if active_pen.isnumeric() else 0)
        self.set_mode(None, GLib.Variant.new_string(config.get('highlight', 'mode')))


    def page_change_action(self, gaction, param):
        """ Change whether we exit or stay in highlighting mode on page changes

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the new mode as a string wrapped in a GLib.Variant
        """
        self.page_change_exits = not gaction.get_state().get_boolean()
        self.config.set('highlight', 'page_change_exits', 'on' if self.page_change_exits else 'off')
        gaction.change_state(GLib.Variant.new_boolean(self.page_change_exits))

        return True


    def set_mode(self, gaction, param):
        """ Change the mode of clearing and restoring highlights

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the new mode as a string wrapped in a GLib.Variant
        """
        new_mode = param.get_string()
        if new_mode not in {'single-page', 'global', 'per-page', 'per-label'}:
            return False

        self.get_application().lookup_action('highlight-mode').change_state(GLib.Variant.new_string(new_mode))
        self.highlight_mode = new_mode
        self.config.set('highlight', 'mode', self.highlight_mode)
        self.remembered_scribbles.clear()

        return True


    def try_cancel(self):
        """ Cancel scribbling, if it is enabled.

        Returns:
            `bool`: `True` if scribbling got cancelled, `False` if it was already disabled.
        """
        if not self.scribbling_mode:
            return False

        self.disable_scribbling()
        return True


    @staticmethod
    def parse_color(text):
        """ Transform a string to a Gdk object in a single function call

        Args:
            text (`str`): A string describing a color

        Returns:
            :class:`~Gdk.RGBA`: A new color object parsed from the string
        """
        color = Gdk.RGBA()
        color.parse(text)
        return color


    def points_to_curves(self, points):
        """ Transform a list of points from scribbles to bezier curves

        Returns:
            `list`: control points of a bezier curves to draw
        """
        curves = []

        if len(points) <= 2:
            return curves

        c1 = points[1]
        for c2, pt in zip(points[2:-1:2], points[3:-1:2]):
            half_c2pt = (pt[0] - c2[0]) / 2, (pt[1] - c2[1]) / 2

            curves.append((*c1, c2[0] + half_c2pt[0], c2[1] + half_c2pt[1], *pt))
            c1 = (pt[0] + half_c2pt[0], pt[1] + half_c2pt[1])

        if len(points) % 2 == 0:
            curves.append((*c1, *points[-2], *points[-1]))

        return curves


    def track_scribble(self, widget, event):
        """ Draw the scribble following the mouse's moves.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        pos = self.get_slide_point(widget, event)
        if self.scribble_drawing:
            self.scribble_list[-1][-1].append(pos)
            self.scribble_redo_list.clear()

            self.adjust_buttons()

        self.mouse_pos = pos
        self.redraw_current_slide()
        return self.scribble_drawing


    def key_event(self, widget, event):
        """ Handle key events to activate the eraser while the shortcut is held

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if not self.scribbling_mode:
            return False
        elif event.type != Gdk.EventType.KEY_PRESS and event.type != Gdk.EventType.KEY_RELEASE:
            return False
        elif not (*event.get_keyval()[1:], event.get_state()) in self.toggle_erase_shortcuts:
            return False

        if event.type == Gdk.EventType.KEY_PRESS and self.active_preset and self.toggle_erase_source is None:
            self.previous_preset = self.active_preset
            self.toggle_erase_source = 'shortcut'
            self.load_preset(target=0)
        elif event.type == Gdk.EventType.KEY_RELEASE and self.toggle_erase_source == 'shortcut' \
                and self.previous_preset and not self.active_preset:
            self.load_preset(target=self.previous_preset)
            self.previous_preset = 0
            self.toggle_erase_source = None
        else:
            return False
        return True


    def toggle_scribble(self, widget, event):
        """ Start/stop drawing scribbles.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if not self.scribbling_mode:
            return False

        if event.get_event_type() == Gdk.EventType.BUTTON_PRESS:
            if any(mod & event.get_state() == mod for mod in self.toggle_erase_modifiers) and self.active_preset \
                    and self.toggle_erase_source is None:
                self.previous_preset = self.active_preset
                self.toggle_erase_source = 'modifier'
                self.load_preset(target=0)

            self.scribble_list.append((self.scribble_color, self.scribble_width, []))
            self.scribble_drawing = True

            return self.track_scribble(widget, event)
        elif event.get_event_type() == Gdk.EventType.BUTTON_RELEASE:
            self.scribble_drawing = False
            self.prerender()

            if not self.active_preset and self.previous_preset and self.toggle_erase_source == 'modifier':
                self.load_preset(target=self.previous_preset)
                self.previous_preset = 0
                self.toggle_erase_source = None

            return True

        return False


    def reset_scribble_cache(self):
        """ Clear the cached scribbles.
        """
        window = self.c_da.get_window()

        if window is None:
            return ValueError('Cannot initialize scribble cache without drawing area window')

        scale = window.get_scale_factor()
        ww, wh = self.c_da.get_allocated_width() * scale, self.c_da.get_allocated_height() * scale
        try:
            self.scribble_cache = window.create_similar_image_surface(cairo.Format.ARGB32, ww, wh, scale)
        except ValueError as e:
            if e.args == ('invalid enum value: 5',):
                # Just try again as this error seems to always happen at the first call but not later on
                # This is the gi introspection of cairo not accepting the value Format.RGB30 (???)
                try:
                    self.scribble_cache = window.create_similar_image_surface(cairo.Format.ARGB32, ww, wh, scale)
                except ValueError:
                    return
            else:
                logger.exception('Error creating highlight cache')
        except cairo.Error:
            logger.warning('Failed creating an ARGB32 surface sized {}x{} scale {} for highlight cache'
                           .format(ww, wh, scale), exc_info=True)
            return
        self.next_render = 0


    def prerender(self):
        """ Commit scribbles to cache so they are faster to draw on the slide
        """
        if self.scribble_cache is None:
            try:
                self.reset_scribble_cache()
            except ValueError as e:
                logger.info(e)
                return

        if self.scribble_cache is None:
            self.next_render = 0
            return

        scale = self.c_da.get_window().get_scale_factor()
        ww, wh = self.scribble_cache.get_width() / scale, self.scribble_cache.get_height() / scale
        pen_scale_factor = max(ww / 900, wh / 900)  # or sqrt of product

        cairo_context = cairo.Context(self.scribble_cache)
        cairo_context.set_line_cap(cairo.LINE_CAP_ROUND)

        draw = slice(self.next_render, -1 if self.scribble_drawing else None)

        for color, width, points in self.scribble_list[draw]:
            self.render_scribble(cairo_context, color, width * pen_scale_factor, [(x * ww, y * wh) for x, y in points])

        del cairo_context

        self.next_render = len(self.scribble_list) + (draw.stop if draw.stop else 0)


    def render_scribble(self, cairo_context, color, width, points):
        """ Draw a single scribble, i.e. a bezier curve, on the cairo context

        Args:
            cairo_context (:class:`~cairo.Context`): The canvas on which to render the drawings
            color (:class:`~Gdk.RGBA`): The color of the scribble
            width (`float`): The width of the curve
            points (`list`): The control points of the curve, scaled to the surface.

        Returns:
            :class:`~cairo.Path`: A copy of the path that was drawn
        """
        if not points:
            return

        # alpha == 0 -> Eraser mode
        cairo_context.set_operator(cairo.OPERATOR_OVER if color.alpha else cairo.OPERATOR_CLEAR)
        cairo_context.set_source_rgba(*color)
        cairo_context.set_line_width(width)

        cairo_context.move_to(*points[0])

        for curve in self.points_to_curves(points):
            cairo_context.curve_to(*curve)

        path = cairo_context.copy_path()

        cairo_context.line_to(*points[-1])
        cairo_context.stroke()

        return path


    def draw_scribble(self, widget, cairo_context):
        """ Perform the drawings by user.

        Args:
            widget (:class:`~Gtk.DrawingArea`): The widget where to draw the scribbles.
            cairo_context (:class:`~cairo.Context`): The canvas on which to render the drawings
        """
        scale = widget.get_window().get_scale_factor()
        ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
        cw, ch = self.scribble_cache.get_width(), self.scribble_cache.get_height()

        cairo_context.save()

        cairo_context.scale(ww * scale / cw, wh * scale / ch)
        cairo_context.set_source_surface(self.scribble_cache)
        cairo_context.paint()

        cairo_context.restore()

        cairo_context.push_group()

        pen_scale_factor = max(ww / 900, wh / 900)  # or sqrt of product
        if self.scribble_drawing:
            cairo_context.set_line_cap(cairo.LINE_CAP_ROUND)
            color, width, points = self.scribble_list[-1]
            self.render_scribble(cairo_context, color, width * pen_scale_factor, [(x * ww, y * wh) for x, y in points])

        cairo_context.pop_group_to_source()
        cairo_context.paint()


        if widget.get_name() == 'scribble_p_da' and self.mouse_pos is not None:
            cairo_context.set_source_rgba(0, 0, 0, 1)
            cairo_context.set_line_width(1)

            mx, my = self.mouse_pos
            cairo_context.arc(mx * ww, my * wh, self.scribble_width * pen_scale_factor / 2, 0, 2 * math.pi)

            cairo_context.stroke_preserve()

            cairo_context.set_source_rgba(*list(self.scribble_color)[:3], self.scribble_color.alpha * .5)
            cairo_context.close_path()
            cairo_context.fill()


    def update_color(self, widget):
        """ Callback for the color chooser button, to set scribbling color.

        Args:
            widget (:class:`~Gtk.ColorButton`):  the clicked button to trigger this event, if any
        """
        self.scribble_color = widget.get_rgba()
        self.update_active_color_width()


    def update_width(self, widget, event, value):
        """ Callback for the width chooser slider, to set scribbling width.

        Args:
            widget (:class:`~Gtk.Scale`): The slider control used to select the scribble width
            event (:class:`~Gdk.Event`):  the GTK event triggering this update.
            value (`int`): the width of the scribbles to be drawn
        """
        self.scribble_width = max(5, min(90, int(value)))
        self.update_active_color_width()


    def update_active_color_width(self):
        """ Update modifications to the active scribble color and width, on the pen button and config object
        """
        if not self.active_preset:
            return

        self.color_width[self.active_preset] = self.scribble_color, self.scribble_width
        self.scribble_preset_buttons[self.active_preset].queue_draw()

        pen = self.active_preset
        self.config.set('highlight', 'color_{}'.format(pen), self.scribble_color.to_string())
        self.config.set('highlight', 'width_{}'.format(pen), str(self.scribble_width))


    def adjust_buttons(self):
        """ Properly enable and disable buttons based on scribblings lists.
        """
        self.scribble_undo.set_sensitive(bool(self.scribble_list))
        self.scribble_clear.set_sensitive(bool(self.scribble_list))
        self.scribble_redo.set_sensitive(bool(self.scribble_redo_list))


    def clear_scribble(self, *args):
        """ Callback for the scribble clear button, to remove all scribbles.
        """
        self.scribble_list.clear()

        self.reset_scribble_cache()
        self.redraw_current_slide()
        self.adjust_buttons()


    def page_change(self, page_number, page_label):
        """ Called when we change pages, to clear or restore scribbles

        Args:
            page_number (`int`): The number of the new page
            page_label (`str`): The label of the new page
        """
        if self.highlight_mode == 'per-page':
            current_page = self.current_page[0]
            new_page = page_number
        elif self.highlight_mode == 'per-label':
            current_page = self.current_page[1]
            new_page = page_label

        # Remember whatever the current mode, to facilitate switching modes
        self.current_page = (page_number, page_label)

        if self.highlight_mode == 'global':
            return
        elif self.highlight_mode == 'single-page':
            return self.clear_scribble()
        else:
            # Now optionally save the current scribbles
            if current_page is not None and self.scribble_list:
                self.remembered_scribbles[current_page] = self.scribble_list.copy()

            self.scribble_list = self.remembered_scribbles.pop(new_page, [])

            self.reset_scribble_cache()
            self.adjust_buttons()
            self.prerender()
            self.redraw_current_slide()


    def pop_scribble(self, *args):
        """ Callback for the scribble undo button, to undo the last scribble.
        """
        if self.scribble_list:
            self.scribble_redo_list.append(self.scribble_list.pop())

        self.adjust_buttons()
        self.reset_scribble_cache()
        self.prerender()
        self.redraw_current_slide()


    def redo_scribble(self, *args):
        """ Callback for the scribble undo button, to undo the last scribble.
        """
        if self.scribble_redo_list:
            self.scribble_list.append(self.scribble_redo_list.pop())

        self.adjust_buttons()
        self.prerender()
        self.redraw_current_slide()


    def on_configure_da(self, widget, event):
        """ Transfer configure resize to the cache.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has been resized
            event (:class:`~Gdk.Event`):  the GTK event, which contains the new dimensions of the widget
        """
        # Don't trust those
        if not event.send_event:
            return

        self.resize_cache(widget.get_name(), event.width, event.height)


    def set_tools_orientation(self, gaction, target):
        """ Changes the orientation of the highlighting tool box.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            target (:class:`~GLib.Variant`): the new orientation to set, as a string wrapped in a GLib.Variant

        Returns:
            `bool`: whether the preset was loaded
        """
        orientation = target.get_string()
        if orientation == self.tools_orientation:
            return False
        elif orientation not in ['horizontal', 'vertical']:
            logger.error('Unexpected highlight-tools orientation {}'.format(orientation))
            return False

        self.tools_orientation = orientation
        self.adjust_tools_orientation()

        gaction.change_state(target)
        self.config.set('layout', 'highlight_tools', self.tools_orientation)


    def adjust_tools_orientation(self):
        """ Actually change the highlight tool elements orientations according to self.tools_orientation
        """
        orientation = Gtk.Orientation.VERTICAL if self.tools_orientation == 'vertical' else Gtk.Orientation.HORIZONTAL
        self.preset_toolbar.set_orientation(orientation)
        self.scribble_toolbar.set_orientation(orientation)
        self.scribble_color_toolbox.set_orientation(orientation)
        self.scribble_width_selector.set_orientation(orientation)

        w, h = sorted(self.scribble_width_selector.get_size_request(), reverse=self.tools_orientation != 'vertical')
        self.scribble_width_selector.set_size_request(w, h)

        # NB the parent container is layed out perpendicularly to its contents
        self.scribble_overlay.set_orientation(Gtk.Orientation.HORIZONTAL if self.tools_orientation == 'vertical' else
                                              Gtk.Orientation.VERTICAL)


    def switch_scribbling(self, gaction, target=None):
        """ Starts the mode where one can read on top of the screen.

        Args:

        Returns:
            `bool`: whether the event was consumed
        """
        if target is not None and target == self.scribbling_mode:
            return False

        # Perform the state toggle
        if self.scribbling_mode:
            return self.disable_scribbling()
        else:
            return self.enable_scribbling()


    def enable_scribbling(self):
        """ Enable the scribbling mode.

        Returns:
            `bool`: whether it was possible to enable (thus if it was not enabled already)
        """
        if self.scribbling_mode:
            return False

        self.scribble_off_render.remove(self.scribble_overlay)
        self.load_layout('highlight')

        self.p_central.queue_draw()
        self.scribble_overlay.queue_draw()

        self.scribbling_mode = True
        self.get_application().lookup_action('highlight').change_state(GLib.Variant.new_boolean(self.scribbling_mode))
        self.pen_action.set_enabled(self.scribbling_mode)

        self.p_central.queue_draw()
        extras.Cursor.set_cursor(self.scribble_p_da, 'invisible')
        return True


    def disable_scribbling(self):
        """ Disable the scribbling mode.

        Returns:
            `bool`: whether it was possible to disable (thus if it was not disabled already)
        """
        if not self.scribbling_mode:
            return False

        self.scribbling_mode = False

        extras.Cursor.set_cursor(self.scribble_p_da, 'default')
        self.load_layout(None)
        self.scribble_off_render.add(self.scribble_overlay)

        self.get_application().lookup_action('highlight').change_state(GLib.Variant.new_boolean(self.scribbling_mode))
        self.pen_action.set_enabled(self.scribbling_mode)

        self.p_central.queue_draw()
        extras.Cursor.set_cursor(self.p_central)
        self.mouse_pos = None

        return True


    def load_preset(self, gaction=None, target=None):
        """ Loads the preset color of a given number or designed by a given widget, as an event handler.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            target (:class:`~GLib.Variant`): the new preset to load, as a string wrapped in a GLib.Variant

        Returns:
            `bool`: whether the preset was loaded
        """
        if type(target) == int:
            self.active_preset = target
        else:
            self.active_preset = int(target.get_string()) if target.get_string() != 'eraser' else 0

        target = str(self.active_preset) if self.active_preset else 'eraser'

        self.config.set('highlight', 'active_pen', target)
        self.pen_action.change_state(GLib.Variant.new_string(target))
        self.scribble_color, self.scribble_width = self.color_width[self.active_preset]

        # Presenter-side setup
        self.scribble_color_selector.set_rgba(self.scribble_color)
        self.scribble_width_selector.set_value(self.scribble_width)
        self.scribble_color_selector.set_sensitive(target != 'eraser')
        self.scribble_width_selector.set_sensitive(target != 'eraser')

        # Re-draw the eraser
        self.scribble_p_da.queue_draw()
        self.c_da.queue_draw()

        return True


    def on_eraser_button_draw(self, widget, cairo_context):
        """ Handle drawing the eraser button.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget to update
            cairo_context (:class:`~cairo.Context`):  the Cairo context (or `None` if called directly)
        """
        cairo_context.push_group()
        scale = widget.get_allocated_height() / self.eraser_surface.get_height()
        cairo_context.scale(scale, scale)

        cairo_context.set_source_surface(self.eraser_surface)
        cairo_context.paint()

        cairo_context.pop_group_to_source()
        cairo_context.paint()


    def on_preset_button_draw(self, widget, cairo_context):
        """ Handle drawing the marker/pencil buttons, with appropriate thickness and color.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget to update
            cairo_context (:class:`~cairo.Context`):  the Cairo context (or `None` if called directly)
        """
        button_number = int(widget.get_name().split('_')[-1])
        color, width = self.color_width[button_number]
        icon, mask = self.marker_surfaces[int((width - 1) / 30)]

        ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
        scale = wh / icon.get_height()

        dw, dh = self.scribble_p_da.get_allocated_width(), self.scribble_p_da.get_allocated_height()
        pen_scale_factor = max(dw / 900, dh / 900)  # or sqrt of product
        width *= pen_scale_factor

        cairo_context.push_group()

        # A line demonstrating the scribble style
        cairo_context.set_source_rgba(*color)
        cairo_context.set_line_width(width)
        cairo_context.move_to(0, wh - width / 2)
        cairo_context.line_to(ww, wh - width / 2)
        cairo_context.stroke()

        cairo_context.set_operator(cairo.OPERATOR_DEST_OUT)

        # Clip the line to the lower triangle
        cairo_context.set_source_rgba(0, 0, 0, 1)
        cairo_context.set_line_width(0)
        cairo_context.move_to(0, 0)
        cairo_context.line_to(0, wh)
        cairo_context.line_to(ww, 0)
        cairo_context.close_path()
        cairo_context.fill()

        # Also clip the colored part of the marker
        cairo_context.scale(scale, scale)
        cairo_context.set_source_surface(mask)
        cairo_context.paint()

        cairo_context.pop_group_to_source()
        cairo_context.paint()


        cairo_context.push_group()

        # Fill with desired color
        cairo_context.set_source_rgba(*color)
        cairo_context.rectangle(0, 0, ww, wh)
        cairo_context.fill()

        # Transform for surfaces
        cairo_context.scale(scale, scale)

        # Clip color to the mask
        cairo_context.set_operator(cairo.OPERATOR_DEST_IN)
        cairo_context.set_source_surface(mask)
        cairo_context.paint()

        # Add the rest of the marker
        cairo_context.set_operator(cairo.OPERATOR_OVER)
        cairo_context.set_source_surface(icon)
        cairo_context.paint()

        cairo_context.pop_group_to_source()
        cairo_context.paint()



# -*- coding: utf-8 -*-
#
#       util.py
#
#       Copyright 2009, 2010 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.util` -- various utility functions
-------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import subprocess
import importlib
import os
import sys
import pathlib

if not getattr(sys, 'frozen', False):
    # doesn’t play too well with cx_Freeze
    try:
        import importlib.resources as importlib_resources
    except ImportError:
        import importlib_resources

IS_POSIX = os.name == 'posix'
IS_MAC_OS = sys.platform == 'darwin'
IS_WINDOWS = os.name == 'nt'


if IS_WINDOWS:
    try:
        import winreg
    except ImportError:
        import _winreg as winreg



def get_pympress_meta():
    """ Get metadata (version, etc) from pympress' __init__.py or git describe.

    Returns:
        `dict`: metadata properties (version, contributors) mapped to their values
    """
    module = importlib.import_module('pympress.__init__')
    info = {'version': module.__version__, 'contributors': module.__author__}

    if getattr(sys, 'frozen', False):
        return info

    git_dir = importlib_resources.files('pympress').joinpath('..', '.git')
    if not git_dir.exists():
        return info

    # Try and get a git describe output in case we are on a dirty/editable version
    try:
        command = 'git --git-dir={} describe --tags --long --dirty'.split()
        command[1] = command[1].format(git_dir)  # after spliting in case path has whitespace

        git_version = subprocess.check_output(command, stderr = subprocess.DEVNULL)

        # answer format is: {last tag}-{commit count since tag}-g{commit sha1 hash}[-dirty]
        tag, count, sha, dirty = (git_version.decode('utf-8') + '-').strip().split('-', 3)
        if count != '0' or dirty:
            info['version'] = '{}+{}@{}'.format(tag.lstrip('v'), count, sha.lstrip('g'))

    except subprocess.CalledProcessError:
        logger.debug('Failed to get git describe output', exc_info = True)

    finally:
        return info


def __get_resource_path(*path_parts):
    """ Return the resource path based on whether its frozen or not.

    Paths parts given should be relative to the pympress package dir.

    Args:
        name (`tuple` of `str`): The directories and filename that constitute the path
        to the resource, relative to the pympress distribution

    Returns:
        :class:`~pathlib.Path`: The path to the resource
    """
    if getattr(sys, 'frozen', False):
        root = pathlib.Path(sys.executable).parent
    else:
        root = importlib_resources.files('pympress')
    return root.joinpath(*path_parts)


def get_locale_dir():
    """ Returns the path to the locale directory.

    Returns:
        :class:`~pathlib.Path`: The path to the locale directory
    """
    return __get_resource_path('share', 'locale')


def get_portable_config():
    """ Returns the path to the configuration file for a portable install (i.e. in the install root).

    Returns:
        :class:`~pathlib.Path`: The path to the portable configuration file.
    """
    return __get_resource_path('pympress.conf')


def get_default_config():
    """ Returns the path to the configuration file containing the defaults.

    Returns:
        :class:`~pathlib.Path`: The path to the portable configuration file.
    """
    return __get_resource_path('share', 'defaults.conf')


def get_user_config():
    """ Returns the path to the configuration file in the user config directory

    Returns:
        :class:`~pathlib.Path`: path to the user configuration file.
    """
    if IS_WINDOWS:
        base_dir = pathlib.Path(os.getenv('APPDATA'))
    elif IS_MAC_OS:
        base_dir = pathlib.Path('~/Library/Preferences').expanduser()
    else:
        base_dir = pathlib.Path(os.getenv('XDG_CONFIG_HOME', '~/.config')).expanduser()
        if not base_dir.exists():
            base_dir.mkdir(parents=True)

    return base_dir.joinpath('pympress' + ('.ini' if IS_WINDOWS else ''))


def load_style_provider(style_provider):
    """ Load the css and in a style provider

    Args:
        style_provider (:class:`~Gtk.CssProvider`): The style provider in which to load CSS

    Returns:
        :class:`~Gtk.CssProvider`: The style provider with CSS loaded
    """
    style_provider.load_from_path(str(__get_resource_path('share', 'css', 'default.css')))
    return style_provider


def get_icon_path(name):
    """ Get the path for an image from pympress' resources

    Args:
        name (`str`): The name of the icon to load

    Returns:
        `str`: The path to the icon to load
    """
    return str(__get_resource_path('share', 'pixmaps', name))


def get_ui_resource_file(name, ext='.glade'):
    """ Load an UI definition file from pympress' resources

    Args:
        name (`str`): The name of the UI to load
        ext (`str`): The extension of the file

    Returns:
        `str`: The full path to the glade file
    """
    return str(__get_resource_path('share', 'xml', name + ext))


def list_icons():
    """ List the icons from pympress' resources.

    Returns:
        `list` of `str`: The paths to the icons in the pixmaps directory
    """
    return list(map(str, __get_resource_path('share', 'pixmaps').glob('pympress-*.png')))


def get_log_path():
    """ Returns the appropriate path to the log file in the user app dirs.

    Returns:
        :class:`~pathlib.Path`: path to the log file.
    """
    if IS_WINDOWS:
        base_dir = pathlib.Path(os.getenv('LOCALAPPDATA', os.getenv('APPDATA')))
    elif IS_MAC_OS:
        base_dir = pathlib.Path('~/Library/Logs').expanduser()
    else:
        base_dir = pathlib.Path(os.getenv('XDG_CACHE_HOME', '~/.cache')).expanduser()

    if not base_dir.exists():
        base_dir.mkdir(parents=True)

    return base_dir.joinpath('pympress.log')


def fileopen(f):
    """ Call the right function to open files, based on the platform.

    Args:
        f (path-like): path to the file to open
    """
    if IS_WINDOWS:
        os.startfile(f)
    elif IS_MAC_OS:
        subprocess.call(['open', str(f)])
    else:
        subprocess.call(['xdg-open', str(f)])


def introspect_flag_value(flags_class, nick, fallback):
    """ Get the value of a flag from its class, given a value’s name (or nick)

    Introspection technique (in particular __flags_values__ dict) inspired from pygtkcompat.
    This is needed because there is no typelib for libgstplayback.

    Args:
        flags_class (a `~type` inheriting from :class:`~Gobject.GFlags`): the flags class to introspect
        nick (`str`): a name or nick of the flag value that should be returned
        fallback (`int`): the documented flag value, if lookup fails
    """
    try:
        flag_values = flags_class.__flags_values__
    except AttributeError:
        return fallback

    for value, flag in flag_values.items():
        if nick in flag.value_nicks or nick in flag.value_names:
            return value
    else:
        return fallback


def hard_set_screensaver(disabled):
    """ Enable or disable the screensaver.

    Args:
        disabled (`bool`):  if `True`, indicates that the screensaver must be disabled; otherwise it will be enabled
    """
    if IS_MAC_OS:
        # On Mac OS X we can use caffeinate to prevent the display from sleeping
        if disabled:
            if hard_set_screensaver.caffeinate_process is None or hard_set_screensaver.caffeinate_process.poll():
                hard_set_screensaver.caffeinate_process = subprocess.Popen(['caffeinate', '-d', '-w', str(os.getpid())])
        else:
            if hard_set_screensaver.caffeinate_process and not hard_set_screensaver.caffeinate_process.poll():
                hard_set_screensaver.caffeinate_process.kill()
                hard_set_screensaver.caffeinate_process.poll()
                hard_set_screensaver.caffeinate_process = None

    elif IS_WINDOWS:
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Control Panel\Desktop', 0,
                                winreg.KEY_QUERY_VALUE | winreg.KEY_SET_VALUE) as key:
                if disabled:
                    value, regtype = winreg.QueryValueEx(key, "ScreenSaveActive")
                    assert regtype == winreg.REG_SZ, 'Unexpected RegType when modifying ScreenSaveActive'
                    hard_set_screensaver.dpms_was_enabled = (value == "1")
                    if hard_set_screensaver.dpms_was_enabled:
                        winreg.SetValueEx(key, "ScreenSaveActive", 0, winreg.REG_SZ, "0")
                elif hard_set_screensaver.dpms_was_enabled:
                    winreg.SetValueEx(key, "ScreenSaveActive", 0, winreg.REG_SZ, "1")
        except (OSError, PermissionError, AssertionError):
            logger.exception(_("access denied when trying to access screen saver settings in registry!"))

    elif IS_POSIX:
        logger.warning(_("Should not require hard enable/disable screensaver on Linux"))

    else:
        logger.warning(_("Unsupported OS: can't enable/disable screensaver"))


#: remember DPMS setting before we change it
hard_set_screensaver.dpms_was_enabled = None
#: A :class:`~subprocess.Popen` object to track the child caffeinate process
hard_set_screensaver.caffeinate_process = None

##
# Local Variables:
# mode: python
# indent-tabs-mode: nil
# py-indent-offset: 4
# fill-column: 80
# end:



# -*- coding: utf-8 -*-
#
#       media_overlays/__init__.py
#
#       Copyright 2018 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

""" Backends for video overlay widgets.
"""

__all__ = ['base', 'gif_backend', 'gst_backend', 'vlc_backend']



#!/usr/bin/env python3
#
#       setup.py
#
#       Copyright 2009 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

""" pympress setup script.

Mostly wrapping logic for freezing (with cx_Freeze for windows builds).
All configuration is in setup.cfg.
"""

import os
import re
import sys
import pathlib
import subprocess
from ctypes.util import find_library
import setuptools

from setuptools import Command
from setuptools.command.build_py import build_py


def find_index_startstring(haystack, needle, start=0, stop=sys.maxsize):
    """ Return the index of the first string in haystack starting with needle, or raise ValueError if none match.
    """
    try:
        return next(n for n, v in enumerate(haystack[start:stop], start) if v.startswith(needle))
    except StopIteration:
        raise ValueError('No string starts with ' + needle)


class GettextBuildCatalog(Command):
    """ Patched build command to generate translations .mo files using gettext’s msgfmt

    This is used for build systems that do not have easy access to Babel
    """
    user_options = [
        ('domain=', 'D', "domains of PO files (space separated list, default 'messages')"),
        ('directory=', 'd', 'path to base directory containing the catalogs'),
        ('use-fuzzy', 'f', 'also include fuzzy translations'),
        ('statistics', None, 'print statistics about translations')
    ]

    def initialize_options(self):
        """ Initialize options
        """
        self.domain = None
        self.directory = None
        self.use_fuzzy = False
        self.statistics = True


    def finalize_options(self):
        """ Finalize options
        """
        assert self.domain is not None and self.directory is not None


    def run(self):
        """ Run msgfmt before running (parent) develop command
        """
        po_wildcard = pathlib.Path(self.directory).glob(str(pathlib.Path('*', 'LC_MESSAGES', self.domain + '.po')))
        for po in po_wildcard:
            print(po)
            mo = po.with_suffix('.mo')

            cmd = ['msgfmt', str(po), '-o', str(mo)]
            if self.use_fuzzy:
                cmd.insert(1, '--use-fuzzy')
            if self.statistics:
                cmd.insert(1, '--statistics')

            subprocess.check_output(cmd)



class BuildWithCatalogs(build_py):
    """ Patched build command to generate translations .mo files using Babel

    This is what we use by default, e.g. when distributing through PyPI
    """
    def run(self):
        """ Run compile_catalog before running (parent) develop command
        """
        try:
            self.distribution.run_command('compile_catalog')
        except Exception as err:
            if err.args == ('no message catalogs found',):
                pass  # Running from a source tarball − compiling already done
            else:
                raise
        build_py.run(self)


# All functions listing resources return a list of pairs: (system path, distribution relative path)
def gtk_resources():
    """ Returns a list of the non-DLL Gtk resources to include in a frozen/binary package.
    """
    include_path = pathlib.Path(find_library('libgtk-3-0')).parent
    include_path = include_path.parent if include_path.name in {'bin', 'lib', 'lib64'} else include_path

    include_files = []
    resources = [
        pathlib.Path('etc'),
        pathlib.Path('lib', 'girepository-1.0'),
        pathlib.Path('lib', 'gtk-3.0'),
        pathlib.Path('lib', 'gdk-pixbuf-2.0'),
        pathlib.Path('share', 'poppler'),
        pathlib.Path('share', 'themes'),
        pathlib.Path('share', 'icons'),
        pathlib.Path('share', 'glib-2.0'),
        pathlib.Path('share', 'xml')
    ]

    for f in resources:
        p = include_path.joinpath(f)
        if p.exists():
            include_files.append((str(p), str(f)))
        else:
            print('WARNING: Can not find {} (at {})'.format(f, p))

    return include_files


def dlls():
    """ Returns a list of all DLL files we need to include, in a frozen/binary package on windows.

    Relies on a hardcoded list tested for the appveyor build setup.
    """
    if os.name != 'nt':
        return []

    libs = 'libatk-1.0-0.dll libbrotlicommon.dll libbrotlidec.dll libcurl-4.dll libdatrie-1.dll \
    libepoxy-0.dll libfribidi-0.dll libgdk-3-0.dll libgdk_pixbuf-2.0-0.dll libgif-7.dll \
    libgio-2.0-0.dll libgirepository-1.0-1.dll libglib-2.0-0.dll libgobject-2.0-0.dll libgtk-3-0.dll \
    libidn2-0.dll libjpeg-8.dll liblcms2-2.dll libnghttp2-14.dll libnspr4.dll libopenjp2-7.dll \
    libpango-1.0-0.dll libpangocairo-1.0-0.dll libpangoft2-1.0-0.dll libpangowin32-1.0-0.dll \
    libplc4.dll libplds4.dll libpoppler-105.dll libpoppler-cpp-0.dll libpoppler-glib-8.dll libpsl-5.dll \
    libpython{0.major}.{0.minor}.dll libstdc++-6.dll libthai-0.dll libtiff-5.dll libunistring-2.dll \
    libwinpthread-1.dll libzstd.dll nss3.dll nssutil3.dll smime3.dll'.format(sys.version_info)
    # these appear superfluous, though unexpectedly so:
    # libcairo-2.dll libcairo-gobject-2.dll libfontconfig-1.dll libfreetype-6.dll libiconv-2.dll
    # libgettextlib-0-19-8-1.dll libgettextpo-0.dll libgettextsrc-0-19-8-1.dll libintl-8.dll libjasper-4.dll

    lib_gtk_dir = pathlib.Path(find_library('libgtk-3-0')).parent

    gdbus = pathlib.Path(find_library('gdbus.exe'))
    include_files = [(str(gdbus), str(pathlib.Path('lib', 'gi', 'gdbus.exe'))), (str(gdbus), 'gdbus.exe')]
    for lib in libs.split():
        path = find_library(lib)
        path = pathlib.Path(path) if path is not None else path
        if path is not None and path.exists():
            include_files.append((str(path), lib))
        else:
            lib = pathlib.Path(lib)
            # Look in other directories?
            for path in lib_gtk_dir.glob(re.sub('-[0-9.]*$', '-*', lib.stem) + lib.suffix):
                include_files.append((str(path), path.name))
                print('WARNING: Can not find library {}, including {} instead'.format(lib, path.name))
            else:
                print('WARNING: Can not find library {}'.format(lib))

    return include_files


def check_cli_arg(val):
    """ Check whether an argument was passed, and clear it from sys.argv

    Returns (bool): whether the arguement was present
    """
    if val in sys.argv[1:]:
        sys.argv.remove(val)
        return True

    return False


def pympress_resources():
    """ Return pympress resources. Only for frozen packages, as this is redundant with package_data.
    """
    share = pathlib.Path('pympress', 'share')
    dirs = [share.joinpath('xml'), share.joinpath('pixmaps'), share.joinpath('css'), share.joinpath('defaults.conf')]
    translations = share.glob(str(pathlib.Path('*', 'LC_MESSAGES', 'pympress.mo')))

    return [(str(f), str(f.relative_to('pympress'))) for f in dirs + list(translations)]


if __name__ == '__main__':

    try:
        from babel.messages.frontend import compile_catalog
    except ImportError:
        compile_catalog = GettextBuildCatalog

    options = {'cmdclass': {
        'build_py': BuildWithCatalogs,
        'compile_catalog': compile_catalog,
    }}

    # subtle tweak: don’t put an install section in installed packages
    with open('README.md', encoding='utf-8') as f:
        readme = f.readlines()

        install_section = find_index_startstring(readme, '# Install')
        next_section = find_index_startstring(readme, '# ', install_section + 1)
        del readme[install_section:next_section]

        options['long_description'] = ''.join(readme)


    # Check whether to create a frozen distribution
    if check_cli_arg('--freeze'):
        print('Using cx_Freeze.setup():', file=sys.stderr)
        from cx_Freeze import setup, Executable

        setup(**{
            **options,
            'options': {
                'build_exe': {
                    'includes': [],
                    'excludes': ['tkinter'],
                    'packages': ['codecs', 'gi', 'vlc', 'watchdog'],
                    'include_files': gtk_resources() + dlls() + pympress_resources(),
                    'silent': True
                },
                'bdist_msi': {
                    'add_to_path': True,
                    'all_users': False,
                    'summary_data': {
                        'comments': 'https://github.com/Cimbali/pympress/',
                        'keywords': 'pdf-viewer, beamer, presenter, slide, projector, pdf-reader, \
                                    presentation, python, poppler, gtk, pygi, vlc',
                    },
                    'upgrade_code': '{5D156784-ED69-49FF-A972-CBAD312187F7}',
                    'install_icon': str(pathlib.Path('pympress', 'share', 'pixmaps', 'pympress.ico')),
                    'extensions': [{
                        'extension': 'pdf',
                        'verb': 'open',
                        'executable': 'pympress-gui.exe',
                        'argument': '"%1"',
                        'mime': 'application/pdf',
                        'context': 'Open with p&ympress',
                    }],
                }
            },
            'executables': [
                Executable(str(pathlib.Path('pympress', '__main__.py')), target_name='pympress-gui.exe',
                           base='Win32GUI', shortcut_dir='ProgramMenuFolder', shortcut_name='pympress',
                           icon=str(pathlib.Path('pympress', 'share', 'pixmaps', 'pympress.ico'))),
                Executable(str(pathlib.Path('pympress', '__main__.py')), target_name='pympress.exe',
                           base='Console', icon=str(pathlib.Path('pympress', 'share', 'pixmaps', 'pympress.ico'))),
            ]
        })
    else:
        # Normal behaviour: use setuptools, load options from setup.cfg
        print('Using setuptools.setup():', file=sys.stderr)

        setuptools_version = tuple(int(n) for n in setuptools.__version__.split('.')[:2])
        # older versions are missing out!
        if setuptools_version >= (30, 5):
            options['data_files'] = [
                ('share/pixmaps/', ['pympress/share/pixmaps/pympress.png']),
                ('share/applications/', ['pympress/share/applications/io.github.pympress.desktop']),
            ]

        setuptools.setup(**options)


##
# Local Variables:
# mode: python
# indent-tabs-mode: nil
# py-indent-offset: 4
# fill-column: 80
# end:



# -*- coding: utf-8 -*-
#
#       builder.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.ui_builder` -- abstract GUI management
-----------------------------------------------------

This module contains the tools to load the graphical user interface of pympress,
building the widgets/objects from XML (glade) files, applying translation "manually"
to avoid dealing with all the mess of C/GNU gettext's bad portability.
"""

import logging
logger = logging.getLogger(__name__)

import copy
from collections import deque

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, GLib, Gio

from pympress import util


class Builder(Gtk.Builder):
    """ GUI builder, inherits from :class:`~Gtk.Builder` to read XML descriptions of GUIs and load them.
    """
    #: `set` of :class:`~Gtk.Widget`s that have been built by the builder, and translated
    __built_widgets = set()

    #: `dict` mapping :class:`~Gtk.Paned` names to a `tuple` of (handler id of the size-allocate signal, remaining
    #: number of times we allow this signal to run), and we run the signal 2 * (depth + 1) for each pane.
    #: This is because size allocation is done bottom-up but each pane sets a top-down constraint.
    pending_pane_resizes = {}

    _glib_type_strings = {
        float: 'd',
        bool: 'b',
        int: 'x',
        str: 's',
    }

    _glib_type_getters = {
        'd': GLib.Variant.get_double,
        'b': GLib.Variant.get_boolean,
        'x': GLib.Variant.get_int64,
        's': GLib.Variant.get_string,
    }

    def __init__(self):
        super(Builder, self).__init__()


    @staticmethod
    def __translate_widget_strings(a_widget):
        """ Calls gettext on all strings we can find in a_widgets.

        Args:
            a_widget (:class:`~GObject.Object`): an object built by the builder, usually a widget
        """
        for str_prop in (prop.name for prop in a_widget.props if prop.value_type == GObject.TYPE_STRING):
            try:
                str_val = getattr(a_widget.props, str_prop)
                if str_val:
                    setattr(a_widget.props, str_prop, _(str_val))
            except TypeError:
                # Thrown when a string property is not readable
                pass

    @staticmethod
    def __recursive_translate_menu(menu):
        """ Calls gettext on all strings we can find in widgets, and recursively on its children.

        Args:
            menu (:class:`~Gio.Menu`): the menu to translate
        """
        menu_items = []
        for n in range(menu.get_n_items()):
            item = Gio.MenuItem.new_from_model(menu, n)

            label = item.get_attribute_value(Gio.MENU_ATTRIBUTE_LABEL, GLib.VariantType('s'))
            if label:
                item.set_label(_(label.get_string()))

            menu_items.append(item)

            link_iter = menu.iterate_item_links(n)
            while link_iter.next():
                Builder.__recursive_translate_menu(link_iter.get_value())

        menu.remove_all()
        for item in menu_items:
            menu.append_item(item)


    def get_callback_handler(self, handler_name):
        """ Returns the handler from its name, searching in target.

        Parse handler names and split on '.' to use recursion.

        Args:
            target (`object`): An object that has a method called `handler_name`
            handler_name (`str`): The name of the function to be connected to a signal

        Returns:
            `function`: A function bound to an object
        """
        target = self

        for attr in handler_name.split('.'):
            try:
                target = getattr(target, attr)
            except AttributeError:
                logger.error('Can not reach target of signal {}.{}()'.format(self, handler_name), exc_info=True)
                return None

        return target


    def signal_connector(self, builder, object, signal_name, handler_name, connect_object, flags, *user_data):
        """ Callback for signal connection. Implements the `~Gtk.BuilderConnectFunc` function interface.

        Args:
            builder (:class:`~pympress.builder.Builder`): The builder, unused
            object (:class:`~GObject.Object`): The object (usually a wiget) that has a signal to be connected
            signal_name (`str`): The name of the signal
            handler_name (`str`): The name of the function to be connected to the signal
            connect_object (:class:`~GObject.Object`): unused
            flags (:class:`~GObject.ConnectFlags`): unused
            user_data (`tuple`): supplementary positional arguments to be passed to the handler
        """
        try:
            handler = self.get_callback_handler(handler_name)
            object.connect(signal_name, handler, *user_data)

        except Exception:
            logger.critical('Impossible to connect signal {} from object {} to handler {}'
                            .format(signal_name, object, handler_name), exc_info = True)


    def connect_signals(self, base_target):
        """ Signal connector connecting to properties of `base_target`, or properties of its properties, etc.

        Args:
            base_target (:class:`~pympress.builder.Builder`): The target object, that has functions to be connected to
            signals loaded in this builder.
        """
        Builder.connect_signals_full(base_target, self.signal_connector)


    def load_ui(self, resource_name, **kwargs):
        """ Loads the UI defined in the file named resource_name using the builder.

        Args:
            resource_name (`str`): the basename of the glade file (without extension), identifying the resource to load.
        """
        self.add_from_file(util.get_ui_resource_file(resource_name, **kwargs))

        # Get all newly built objects
        new_objects = set(self.get_objects()) - self.__built_widgets
        self.__built_widgets.update(new_objects)

        for obj in new_objects:
            # pass new objects to manual translation
            if isinstance(obj, Gio.Menu):
                self.__recursive_translate_menu(obj)
            else:
                self.__translate_widget_strings(obj)

            # Instrospectively load objects. If we have a self.attr == None and this attr is the name of a built object,
            # link it together.
            if issubclass(type(obj), Gtk.Buildable):
                obj_id = Gtk.Buildable.get_name(obj)

                # set Gtk.Widget.name property to the value of the Gtk.Buildable id
                if issubclass(type(obj), Gtk.Widget):
                    Gtk.Widget.set_name(obj, obj_id)

                if hasattr(self, obj_id) and getattr(self, obj_id) is None:
                    setattr(self, obj_id, obj)


    def list_attributes(self, target):
        """ List the None-valued attributes of target.

        Args:
            target (`dict`): An object with None-valued attributes
        """
        for attr in dir(target):
            try:
                if attr[:2] + attr[-2:] != '____' and getattr(target, attr) is None:
                    yield attr
            except RuntimeError:
                pass


    def load_widgets(self, target):
        """ Fill in target with the missing elements introspectively.

        This means that all attributes of `target` that are None now must exist under the same name in the builder.

        Args:
            target (`dict`): An object with None-valued properties whose names correspond to ids of built widgets.
        """
        for attr in self.list_attributes(target):
            setattr(target, attr, self.get_object(attr))


    def replace_layout(self, layout, top_widget, leaf_widgets, pane_resize_handler=None):
        """ Remix the layout below top_widget with the layout configuration given in 'layout' (assumed to be valid!).

        Args:
            layout (`dict`): the json-parsed config string, thus a hierarchy of lists/dicts, with strings as leaves
            top_widget (:class:`~Gtk.Container`): The top-level widget under which we build the hierachyy
            leaf_widgets (`dict`): the map of valid leaf identifiers (strings) to the corresponding :class:`~Gtk.Widget`
            pane_resize_handler (function): callback function to be called when the panes are resized

        Returns:
            `dict`: The mapping of the used :class:`~Gtk.Paned` widgets to their relative handle position (in 0..1).
        """
        # take apart the previous/default layout
        containers = []
        widgets = top_widget.get_children()
        i = 0
        while i < len(widgets):
            w = widgets[i]
            if w in self.placeable_widgets.values():
                pass
            elif issubclass(type(w), Gtk.Box) or issubclass(type(w), Gtk.Paned):
                widgets.extend(w.get_children())
                containers.append(w)
            w.get_parent().remove(w)
            i += 1

        # cleanup widgets
        del widgets[:]
        while containers:
            containers.pop().destroy()
        self.pending_pane_resizes.clear()

        # iterate over new layout to build it, using a BFS
        widgets_to_add = deque([(top_widget, copy.deepcopy(layout))])
        pane_resize = set()
        pane_handle_pos = {}

        while widgets_to_add:
            parent, w_desc = widgets_to_add.popleft()

            if type(w_desc) is str:
                w = leaf_widgets[w_desc]

            else:
                # get new container widget
                if 'resizeable' in w_desc and w_desc['resizeable']:
                    orientation = getattr(Gtk.Orientation, w_desc['orientation'].upper())
                    w = Gtk.Paned.new(orientation)
                    w.set_wide_handle(True)

                    # Add on resize events
                    if pane_resize_handler:
                        w.connect('notify::position', pane_resize_handler)
                        w.connect('button-release-event', pane_resize_handler)

                    # left pane is first child
                    widgets_to_add.append((w, w_desc['children'].pop()))

                    if 'proportions' in w_desc:
                        right_pane = w_desc['proportions'].pop()
                        left_pane  = w_desc['proportions'].pop()
                        w_desc['proportions'].append(left_pane + right_pane)

                        pane_handle_pos[w] = left_pane / (left_pane + right_pane)
                        pane_resize.add(w)
                    else:
                        pane_handle_pos[w] = 0.5

                    hid = w.connect('size-allocate', self.resize_paned, pane_handle_pos[w])

                    w.set_name('GtkPaned{}'.format(len(self.pending_pane_resizes)))
                    self.pending_pane_resizes[w.get_name()] = (hid, 2 * len(self.pending_pane_resizes) + 1)

                    # if more than 2 children are to be added, add the 2+ from the right side in a new child Gtk.Paned
                    widgets_to_add.append((w, w_desc['children'][0] if len(w_desc['children']) == 1 else w_desc))
                else:
                    w = Gtk.Box.new(getattr(Gtk.Orientation, w_desc['orientation'].upper()), 5)
                    w.set_homogeneous(True)
                    w.set_spacing(10)

                    widgets_to_add.extend((w, c) for c in w_desc['children'])

            if issubclass(type(parent), Gtk.Box):
                parent.pack_start(w, True, True, 0)
            else:
                # it's a Gtk.Paned
                if parent.get_child2() is None:
                    parent.pack2(w, True, True)
                    if parent.get_orientation() == Gtk.Orientation.HORIZONTAL:
                        w.set_margin_start(8)
                    else:
                        w.set_margin_top(8)
                else:
                    parent.pack1(w, True, True)
                    if parent.get_orientation() == Gtk.Orientation.HORIZONTAL:
                        w.set_margin_end(8)
                    else:
                        w.set_margin_bottom(8)

        return pane_handle_pos


    def resize_paned(self, paned, rect, relpos):
        """ Resize `paned` to have its handle at `relpos`, then disconnect this signal handler.

        Called from the :func:`Gtk.Widget.signals.size_allocate` signal.

        Args:
            paned (:class:`~Gtk.Paned`): Panel whose size has just been allocated, and whose handle needs initial
                                         placement.
            rect (:class:`~Gdk.Rectangle`): The rectangle specifying the size that has just been allocated to `~paned`
            relpos (`float`): A number between `0.` and `1.` that specifies the handle position

        Returns:
            `True`
        """
        size = rect.width if paned.get_orientation() == Gtk.Orientation.HORIZONTAL else rect.height
        handle_pos = int(round(relpos * size))
        GLib.idle_add(paned.set_position, handle_pos)

        name = paned.get_name()
        handler_id, sizes = self.pending_pane_resizes[name]
        self.pending_pane_resizes[name] = (handler_id, sizes - 1)

        if sizes <= 0:
            paned.disconnect(handler_id)

        return True


    @staticmethod
    def setup_actions(actions, action_map=None):
        """ Sets up actions with a given prefix, using the Application as the ActionMap.

        Args:
            actions (`dict`): Maps the action names to dictionaries containing their parameters.
            action_map (:class:`~Gio.ActionMap`): The object implementing the action map interface to register actions
        """
        if action_map is None:
            action_map = Gio.Application.get_default()

        for action_name, details in actions.items():
            state = details.get('state')
            param = details.get('parameter_type')
            enabled = details.get('enabled')

            if param is not None:
                param = GLib.VariantType.new(Builder._glib_type_strings[param])

            if state is not None:
                state = GLib.Variant(Builder._glib_type_strings[type(state)], state)
                action = Gio.SimpleAction.new_stateful(action_name, param, state)
            else:
                action = Gio.SimpleAction.new(action_name, param)

            for event in ['activate', 'change-state']:
                handler = details.get(event.replace('-', '_'))
                if handler is not None:
                    action.connect(event, handler)

            action_map.add_action(action)

            if enabled is not None:
                action.set_enabled(enabled)

        return action_map



# -*- coding: utf-8 -*-
#
#       media_overlays/gif.py
#
#       Copyright 2018 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
:mod:`pympress.media_overlays.gif` -- widget to play gif images as videos
-------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import gi
import cairo
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, GLib, GdkPixbuf


from pympress.media_overlays import base


class GifOverlay(base.VideoOverlay):
    """ A simple overlay mimicking the functionality of showing videos, but showing gifs instead.
    """
    #: A :class:`~GdkPixbuf.PixbufAnimation` containing all the frames and their timing for the displayed gif
    anim = None
    #: A :class:`~GdkPixbuf.PixbufAnimationIter` which will provide the timely access to the frames in `~anim`
    iter = None
    #: A `tuple` of (`int`, `int`) indicating the size of the bounding box of the gif
    base_size = None
    #: The :class:`~cairo.Matrix` defining the zoom & shift to scale the gif
    transform = None

    def __init__(self, *args, **kwargs):
        super(GifOverlay, self).__init__(*args, **kwargs)

        # override: no toolbar or interactive stuff for a gif, replace the whole widget area with a GdkPixbuf
        self.autoplay = True
        self.toolbar.set_visible(False)

        # we'll manually draw on the movie zone
        self.movie_zone.connect('draw', self.draw)
        self.movie_zone.connect('configure-event', self.set_transform)


    def _set_file(self, filepath):
        """ Sets the media file to be played by the widget.

        Args:
            filepath (`pathlib.Path`): The path to the media file path
        """
        self.anim = GdkPixbuf.PixbufAnimation.new_from_file(str(filepath))
        self.base_size = (self.anim.get_width(), self.anim.get_height())
        self.iter = self.anim.get_iter(None)

        self.set_transform()
        self.advance_gif()


    def set_transform(self, *args):
        """ Compute the transform to scale (not stretch nor crop) the gif.
        """
        widget_size = (self.movie_zone.get_allocated_width(), self.movie_zone.get_allocated_height())
        scale = min(widget_size[0] / self.base_size[0], widget_size[1] / self.base_size[1])
        dx = widget_size[0] - scale * self.base_size[0]
        dy = widget_size[1] - scale * self.base_size[1]

        self.transform = cairo.Matrix(xx = scale, yy = scale, x0 = dx / 2, y0 = dy / 2)


    def draw(self, widget, ctx):
        """ Simple resized drawing: get the pixbuf, set the transform, draw the image.
        """
        if self.iter is None:
            return False

        try:
            ctx.transform(self.transform)
            Gdk.cairo_set_source_pixbuf(ctx, self.iter.get_pixbuf(), 0, 0)
            ctx.paint()
        except cairo.Error:
            logger.error(_('Cairo can not draw gif'), exc_info = True)


    def advance_gif(self):
        """ Advance the gif, queue redrawing if the frame changed, and schedule the next frame.
        """
        if self.iter.advance():
            self.movie_zone.queue_draw()

        delay = self.iter.get_delay_time()
        if delay >= 0:
            GLib.timeout_add(delay, self.advance_gif)


    def do_set_time(self, t):
        """ Set the player at time t.

        Should run on the main thread to ensure we avoid reentrency problems.

        Args:
            t (`int`): the timestamp, in ms

        Returns:
            `bool`: `True` iff this function should be run again (:meth:`~GLib.idle_add` convention)
        """
        start = GLib.TimeVal()
        GLib.DateTime.new_now_local().to_timeval(start)
        start.add(-t)
        self.iter = self.anim.get_iter(start)
        self.advance_gif()
        return False


    # a bunch of inherited functions that do nothing, for gifs
    def mute(self, *args): pass
    def is_playing(self): return True
    def do_stop(self): pass
    def do_play(self): return False
    def do_play_pause(self): return False

    @classmethod
    def setup_backend(cls):
        """ Returns the name of this backend.
        """
        return _('GdkPixbuf gif player')



# -*- coding: utf-8 -*-
#
#       config.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.config` -- Configuration
---------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import pathlib
import json
import configparser
from collections import deque

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gio

from pympress import util


class Config(configparser.ConfigParser, object):  # python 2 fix
    """ Manage configuration :Get the configuration from its file and store its back.
    """
    #: `dict`-tree of presenter layouts for various modes
    layout = {}
    #: `dict`-tree of presenter layouts that are not configurable
    static_layout = {
        'deck-overview': 'deck',
    }

    #: `dict` of strings that are the valid representations of widgets from the presenter window
    #: that can be dynamically rearranged, mapping to their names
    placeable_widgets = {"notes": "p_frame_notes", "current": "p_frame_cur", "next": "grid_next",
                         "annotations": "p_frame_annot", "highlight": "scribble_overlay"}

    #: `dict` mapping layout ids to tuples of their expected and optional widgets
    widget_reqs = {
        'notes':      (set(placeable_widgets.keys()) - {"annotations", "highlight"}, {"annotations"}),
        'plain':      (set(placeable_widgets.keys()) - {"notes", "highlight"},),
        'note_pages': (set(placeable_widgets.keys()) - {"current", "highlight"},),
        'highlight':  ({"highlight"}, set(placeable_widgets.keys()) - {"highlight"}),
        'highlight_notes':  ({"highlight"}, set(placeable_widgets.keys()) - {"highlight"}),
    }

    #: `dict` mapping accelerator keys to actions
    shortcuts = {}

    @staticmethod
    def path_to_config(search_legacy_locations=False):
        """ Return the path to the currently used configuration file.

        Args:
            search_legacy_locations (`bool`): whether to look in previously used locations

        Returns:
            :class:`~pathlib.Path`: The path to the config file to use
        """
        portable_config = util.get_portable_config()
        if portable_config.exists():
            return portable_config

        user_config = util.get_user_config()

        # migrate old configuration files from previously-used erroneous locations
        if search_legacy_locations and (util.IS_POSIX or util.IS_MAC_OS) and not user_config.exists():
            for legacy_location in ['~/.pympress', '~/.config/pympress']:
                legacy_location = pathlib.Path(legacy_location).expanduser()
                if legacy_location.exists():
                    legacy_location.rename(user_config)

        return user_config


    @staticmethod
    def toggle_portable_config(gaction, param=None):
        """ Create or remove a configuration file for portable installs.

        The portable install file will be used by default, and deleting it causes the config
        to fall back to the user profile location.

        No need to populate the new config file, this will be done on pympress exit.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        portable_config = util.get_portable_config()
        if Config.using_portable_config():
            portable_config.unlink()
        else:
            with open(portable_config, 'w'):
                pass
        gaction.set_state(GLib.Variant.new_boolean(Config.using_portable_config()))


    @staticmethod
    def using_portable_config():
        """ Checks which configuration file location is in use.

        Returns:
            `bool`: `True` iff we are using the portable (i.e. in install dir) location
        """
        return util.get_portable_config().exists()


    def __init__(config):
        # Remove : from delimiters so we can use it in preferences
        super(Config, config).__init__(delimiters=('=',))

        # populate values first from the default config file, then from the proper one
        config.read(util.get_default_config())
        config.load_window_layouts()

        all_commands = dict(config.items('shortcuts')).keys()

        config.read(config.path_to_config(True))
        config.upgrade()
        config.load_window_layouts()

        for command in all_commands:
            # NB only parsing commands from defaults
            parse_ok, action_name, target_value = Gio.Action.parse_detailed_name('app.' + command)
            if not parse_ok or not Gio.action_name_is_valid(action_name):
                logger.error('Failed parsing command ' + command)
                continue

            parsed_accels = {keys: Gtk.accelerator_parse(keys) for keys in config.get('shortcuts', command).split()}
            failed = [keys for keys, parsed in parsed_accels.items() if parsed == (0, 0)]
            if failed:
                logger.warning('Failed parsing shortcut(s) for "{}": "{}"'.format(command, '", "'.join(failed)))

            keep_accels = [keys for keys, parsed in parsed_accels.items() if parsed != (0, 0)]
            if keep_accels:
                config.shortcuts[command] = keep_accels


    def register_actions(self, builder):
        """ Register actions that impact the config file only.

        Args:
            builder (:class:`pympress.builder.Builder`): a builder to setup the actions
        """
        p_full = self.getboolean('presenter', 'start_fullscreen')
        c_full = self.getboolean('content', 'start_fullscreen')
        blank = self.getboolean('content', 'start_blanked')
        portable = self.using_portable_config()

        builder.setup_actions({
            'start-content-fullscreen':   dict(activate=self.toggle_start, state=c_full),
            'start-presenter-fullscreen': dict(activate=self.toggle_start, state=p_full),
            'start-blanked':              dict(activate=self.toggle_start, state=blank),
            'portable-config':            dict(activate=self.toggle_portable_config, state=portable),
        })


    def upgrade(self):
        """ Update obsolete config options when pympress updates.
        """
        if self.get('presenter', 'pointer') == 'pointer_none':
            self.set('presenter', 'pointer', 'red')
            self.set('presenter', 'pointer_mode', 'disabled')

        if self.has_section('scribble') and self.has_option('scribble', 'color'):
            self.set('scribble', 'color_9', self.get('scribble', 'color'))
            self.remove_option('scribble', 'color')
            self.set('scribble', 'active_pen', '9')

        if self.has_section('scribble') and self.has_option('scribble', 'width'):
            self.set('scribble', 'width_9', self.get('scribble', 'width'))
            self.remove_option('scribble', 'width')
            self.set('scribble', 'active_pen', '9')

        if self.has_option('presenter', 'monitor'):
            self.remove_option('presenter', 'monitor')

        if self.has_option('content', 'monitor'):
            self.remove_option('content', 'monitor')

        if self.has_section('scribble'):
            for key, val in self.items('scribble'):
                self.set('highlight', key, val)
            self.remove_section('scribble')

        if self.has_section('gst'):
            for key, val in self.items('gst'):
                self.set('gstreamer', key, 'on' if key == 'enabled' else val)
            self.remove_section('gst')

        # When we went from gtk signal handlers to actions, some renaming had to be done
        for old, new in {
            'next':                 'next-page',
            'prev':                 'prev-page',
            'next_label':           'next-label',
            'prev_label':           'prev-label',
            'hist_back':            'hist-back',
            'hist_forward':         'hist-forward',
            'first':                'first-page',
            'last':                 'last-page',
            'goto_page':            'goto-page',
            'jumpto_label':         'jumpto-label',
            'fullscreen_content':   'content-fullscreen',
            'fullscreen_presenter': 'presenter-fullscreen',
            'pause_timer':          'pause-timer',
            'reset_timer':          'reset-timer',
            'talk_time':            'edit-talk-time',
            'blank_screen':         'blank-screen',
            'notes_mode':           'notes-mode',
            'swap_screens':         'swap-screens',
            'open_file':            'pick-file',
            'close_file':           'close-file',
            'validate':             'validate-input',
            'cancel':               'cancel-input',
            'undo_scribble':        'highlight-undo',
            'redo_scribble':        'highlight-redo',
            'scribble_preset_1':    'highlight-use-pen::1',
            'scribble_preset_2':    'highlight-use-pen::2',
            'scribble_preset_3':    'highlight-use-pen::3',
            'scribble_preset_4':    'highlight-use-pen::4',
            'scribble_preset_5':    'highlight-use-pen::5',
            'scribble_preset_6':    'highlight-use-pen::6',
            'scribble_preset_7':    'highlight-use-pen::7',
            'scribble_preset_8':    'highlight-use-pen::8',
            'scribble_preset_9':    'highlight-use-pen::9',
            'scribble_preset_0':    'highlight-use-pen::eraser',
            'toggle_pointermode':   'pointer-mode::toggle',
        }.items():
            shortcut = self.get('shortcuts', old, fallback=None)
            if shortcut is not None:
                if old in {'hist_back', 'hist_forward'} and 'backspace' in shortcut.lower():
                    # In the Gio.Action operations, accelerators have precedence over widgets, which
                    # means it’s very annoying to edit a shortcut if backspace is mapped to anything.
                    logger.warning('Changing shortcut for "{}" to new default instead of keeping backspace'.format(new))
                else:
                    self.set('shortcuts', new, shortcut)

                self.remove_option('shortcuts', old)


    def getlist(self, *args):
        """ Parse a config value and return the list by splitting the value on commas.

        i.e. bar = foo,qux  returns the list ['foo', 'qux']

        Returns:
            `list`: a config value split into a list.
        """
        return [t.strip() for t in self.get(*args).split(',') if t.strip()]


    def getint(self, *args, **kwargs):
        """ Wrapper for configparser’s getint to handle parsing errors when a fallback is given.

        See :meth:`~configparser.Configparser.getint()`
        """
        try:
            return super(Config, self).getint(*args, **kwargs)
        except ValueError:
            if 'fallback' not in kwargs:
                raise

            logger.warning(_('Error parsing option from config file {}.{} "{}" to int'.format(*args, self.get(*args))),
                           exc_info=True)
            return kwargs['fallback']


    def getfloat(self, *args, **kwargs):
        """ Wrapper for configparser’s to handle parsing errors when a fallback is given.

        See :meth:`~configparser.Configparser.getfloat()`
        """
        try:
            return super(Config, self).getfloat(*args, **kwargs)
        except ValueError:
            if 'fallback' not in kwargs:
                raise

            logger.warning(_('Error parsing option from config file {}.{} "{}" to float')
                           .format(*args, self.get(*args)), exc_info=True)
            return kwargs['fallback']


    def getboolean(self, *args, **kwargs):
        """ Wrapper for configparser’s getboolean to handle parsing errors when a fallback is given.

        :meth:`~configparser.Configparser.getboolean()`
        """
        try:
            return super(Config, self).getboolean(*args, **kwargs)
        except ValueError:
            if 'fallback' not in kwargs:
                raise

            logger.warning(_('Error parsing option from config file {}.{} "{}" to bool'.format(*args, self.get(*args))),
                           exc_info=True)
            return kwargs['fallback']


    def save_config(self):
        """ Save the configuration to its file.
        """
        # serialize the layouts
        for layout_name in self.layout:
            self.set('layout', layout_name, json.dumps(self.layout[layout_name], indent=4))

        with open(self.path_to_config(), 'w') as configfile:
            self.write(configfile)


    def toggle_start(self, gaction, param=None):
        """ Generic function to toggle some boolean startup configuration.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        # action is named start(-presenter|-content)?-property
        start, *win, prop = gaction.get_name().split('-')
        window = win[0] if win else 'content'
        start_conf = 'start_' + prop

        new_state = not gaction.get_state().get_boolean()

        gaction.set_state(GLib.Variant.new_boolean(new_state))
        self.set(window, start_conf, 'on' if new_state else 'off')


    def validate_layout(self, layout, expected_widgets, optional_widgets = set()):
        """ Validate layout: check whether the layout of widgets built from the config string is valid.

        Args:
            layout (`dict`): the json-parsed config string
            expected_widgets (`set`): strings with the names of widgets that have to be used in this layout
            optional_widgets (`set`): strings with the names of widgets that may or may not be used in this layout


        Layout must have all self.placeable_widgets (leaves of the tree, as `str`) and only allowed properties
        on the nodes of the tree (as `dict`).

        Constraints on the only allowed properties of the nodes are:
        - resizeable: `bool` (optional, defaults to no),
        - orientation: `str`, either "vertical" or "horizontal" (mandatory)
        - children: `list` of size >= 2, containing `str`s or `dict`s (mandatory)
        - proportions: `list` of `float` with sum = 1, length == len(children), representing the relative sizes
        of all the resizeable items (if and only if resizeable).

        """
        next_visits = deque([layout])
        widget_seen = set()
        while next_visits:
            w_desc = next_visits.popleft()
            if type(w_desc) is str:
                if w_desc not in expected_widgets and w_desc not in optional_widgets:
                    raise ValueError('Unrecognized widget "{}", pick one of: {}'
                                     .format(w_desc, ', '.join(expected_widgets)))
                elif w_desc in widget_seen:
                    raise ValueError('Duplicate widget "{}", all expected_widgets can only appear once'.format(w_desc))
                widget_seen.add(w_desc)

            elif type(w_desc) is dict:
                if 'orientation' not in w_desc or w_desc['orientation'] not in ['horizontal', 'vertical']:
                    raise ValueError('"orientation" is mandatory and must be "horizontal" or "vertical" at node {}'
                                     .format(w_desc))
                elif 'children' not in w_desc or type(w_desc['children']) is not list or len(w_desc['children']) < 2:
                    raise ValueError('"children" is mandatory and must be a list of 2+ items at node {}'.format(w_desc))
                elif 'resizeable' in w_desc and type(w_desc['resizeable']) is not bool:
                    raise ValueError('"resizeable" must be boolean at node {}'.format(w_desc))

                elif 'proportions' in w_desc:
                    if 'resizeable' not in w_desc or not w_desc['resizeable']:
                        raise ValueError('"proportions" is only valid for resizeable widgets at node {}'.format(w_desc))
                    elif type(w_desc['proportions']) is not list or \
                            any(type(n) is not float for n in w_desc['proportions']) or \
                            len(w_desc['proportions']) != len(w_desc['children']) or \
                            sum(w_desc['proportions']) < 1e-2:
                        raise ValueError('"proportions" must be a list of floats (one per separator), '
                                         'that sum to 1, at node {}'.format(w_desc))
                    elif abs(sum(w_desc['proportions']) - 1) > 1e-10:
                        w_desc['proportions'] = [p / sum(w_desc['proportions']) for p in w_desc['proportions']]


                next_visits.extend(w_desc['children'])
            else:
                raise ValueError('Unexpected type {}, nodes must be dicts or strings, at node {}'
                                 .format(type(w_desc), w_desc))
        widget_missing = expected_widgets - widget_seen
        if widget_missing:
            raise ValueError('Following placeable_widgets were not specified: {}'.format(', '.join(widget_missing)))


    def update_layout_tree(self, layout_name, layout):
        """ Update the layout named `~layout_name`. Throws ValueError on invalid layouts.

        Args:
            layout_name (`str`): the name of the layout to update
            layout (`dict`): the hierarchy of dictionaries, lists, and strings representing the layout
        """
        self.validate_layout(layout, *self.widget_reqs[layout_name])
        self.layout[layout_name] = layout


    def load_window_layouts(self):
        """ Parse and validate layouts loaded from config, with fallbacks if needed.
        """
        for layout_name in self.widget_reqs:
            # Log error and keep default layout
            try:
                self.update_layout_tree(layout_name, json.loads(self.get('layout', layout_name)))
            except ValueError:
                logger.exception('Invalid layout for {}'.format(layout_name))


    def widget_layout_to_tree(self, widget, pane_handle_pos):
        """ Build a tree representing a widget hierarchy, leaves are strings and nodes are `dict`.

        Recursive function. See validate_layout() for more info on the tree structure.

        Args:
            widget (:class:`~Gtk.Widget`): the widget where to start
            pane_handle_pos (`dict`): Map of :class:`~Gtk.Paned` to the relative handle position (float in 0..1)

        Returns:
            `dict`: A tree of dicts reprensenting the widget hierarchy
        """
        orientation_names = {Gtk.Orientation.HORIZONTAL: 'horizontal', Gtk.Orientation.VERTICAL: 'vertical'}

        name = widget.get_name()
        matching_widget_names = [k for k, v in self.placeable_widgets.items() if v == name]

        if matching_widget_names:
            return matching_widget_names[0]

        elif issubclass(type(widget), Gtk.Box):
            return {'resizeable': False, 'orientation': orientation_names[widget.get_orientation()],
                    'children': [self.widget_layout_to_tree(c, pane_handle_pos) for c in widget.get_children()]}

        elif issubclass(type(widget), Gtk.Paned):
            proportions = [1]
            reverse_children = []
            orientation = widget.get_orientation()
            if orientation == Gtk.Orientation.HORIZONTAL:
                get_size = Gtk.Widget.get_allocated_width
            else:
                get_size = Gtk.Widget.get_allocated_height

            while issubclass(type(widget), Gtk.Paned) and orientation == widget.get_orientation():
                left_pane = widget.get_child1()
                right_pane = widget.get_child2()

                visible = left_pane.get_visible() and right_pane.get_visible()
                position = widget.get_position()
                widget_size = get_size(widget)

                if not visible or widget_size <= 1:
                    # reuse number that was in config initially, otherwise gets overwritten with 0
                    ratio = pane_handle_pos[widget]
                else:
                    ratio = float(position) / widget_size
                    pane_handle_pos[widget] = ratio

                proportions = [ratio] + [(1 - ratio) * p for p in proportions]
                reverse_children.append(right_pane)
                widget = left_pane

            reverse_children.append(left_pane)

            return {'resizeable': True, 'proportions': proportions, 'orientation': orientation_names[orientation],
                    'children': [self.widget_layout_to_tree(c, pane_handle_pos) for c in reversed(reverse_children)]}

        raise ValueError('Error serializing layout: widget of type {} '.format(type(widget)) +
                         'is not an expected container or named widget: "{}"'.format(name))


    def get_layout(self, layout_name):
        """ Getter for the `~layout_name` layout.
        """
        try:
            return self.layout[layout_name]
        except KeyError:
            return self.static_layout[layout_name]


    def update_layout_from_widgets(self, layout_name, widget, pane_handle_pos):
        """ Setter for the notes layout.

        Args:
            layout_name (`str`): the name of the layout to update
            widget (:class:`~Gtk.Widget`): the widget that will contain the layout.
            pane_handle_pos (`dict`): Map of :class:`~Gtk.Paned` to the relative handle position (float in 0..1)
        """
        if layout_name not in self.static_layout:
            return
        self.update_layout_tree(layout_name, self.widget_layout_to_tree(widget, pane_handle_pos))
